class SpecConfig:
    META = '.meta'
    TABLE = 'table-name'

    TYPE = 'type'
    COLUMN = 'column'

    ID = 'id'
    SEQUENCE = 'sequence'

    RELATION = 'relation'
    JOIN_TABLE = 'join-table'
    JOIN_COLUMN = 'join-column'
    INVERSE_JOIN_COLUMN = 'inverse-join-column'
    MAPPED_BY = 'mapped-by'



from pathlib import Path
from typing import Dict, List

from crudhex.domain.models import Entity, Field
from crudhex.domain.utils.class_type_utils import get_field_imports, get_field_types
from ..config_context import get_config

from crudhex.domain.ports import db_code_writer

DB_ENTITY_SUFFIX = 'Entity'


def create_entity_class(entity: Entity, folder: Path) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_entity_type_name(entity)
    entity_file = folder / f'{class_type}.java'

    db_code_writer.create_entity(
        entity_file, class_type, get_package(),
        _get_entity_imports(entity), _get_entity_meta(entity), _get_entity_fields_data(entity)
    )

    return entity_file


def get_package() -> str:
    return get_config().db_models_pkg


def get_entity_type_name(entity: Entity) -> str:
    return f'{entity.name}{DB_ENTITY_SUFFIX}'


def _get_entity_meta(entity: Entity) -> Dict[str, str]:
    meta = entity.meta
    if not meta:
        meta = Entity.Meta.default(entity.name)

    return meta.to_dict()


def _get_entity_imports(entity: Entity) -> List[str]:
    imports = []
    for field in entity.fields:
        imports += get_field_imports(field)

    return imports


def _get_entity_fields_data(entity: Entity) -> List[Dict[str, str]]:
    fields_data = []
    for field in entity.fields:
        fields_data.append(_get_field_data(field))

    return fields_data


def _get_field_data(field: Field) -> Dict[str, str]:
    field_data = {
        'name': field.name,
        'column_name': field.column,
        'relationship': False,
        'id': False,
    }
    field_data.update(get_field_types(field)._asdict())

    if field.is_id():
        field_data['id'] = True
        field_data.update(**field.id_meta.to_dict())

    elif field.has_relation():
        field_data['relationship'] = field.relation.type
        field_data.update(**field.relation.to_dict())

    return field_data



from .field import Field
from typing import List, Optional


class Entity:
    name: str
    meta: Optional['Entity.Meta'] = None
    fields: List[Field]

    def __init__(self, name: str):
        self.name = name
        self.fields = []

    def has_meta(self):
        return self.meta is not None

    def get_id_field(self) -> Optional[Field]:
        for field in self.fields:
            if field.is_id(): return field

        return None

    class Meta:
        table_name: str

        def __init__(self, table_name: str):
            self.table_name = table_name

        def to_dict(self):
            return self.__dict__

        @classmethod
        def default(cls, entity_name: str) -> 'Entity.Meta':
            return cls(entity_name.lower())



from pathlib import Path
from typing import Optional

import typer
from rich.console import Console
from rich.theme import Theme
from rich.progress import Progress

from crudhex.domain.services import dsl_parser, config_context, db_adapter_generator, domain_generator


_CONF_HELP = f'Project config file to know packages and code paths. defaults to {config_context.DEFAULT_CONFIG}'
_SPEC_HELP = 'Spec file path to process'

app = typer.Typer()
out_console: Optional[Console] = None
err_console: Optional[Console] = None


def main():
    _setup()
    app()


@app.command()
def generate(
        project_config: str = typer.Option(None, '--config', '-c', help=_CONF_HELP),
        spec_file: str = typer.Argument(..., help=_SPEC_HELP)
):
    if not project_config:
        out_console.print('Loading from default config path...', style='info')

    config_context.load_config(Path(project_config) if project_config else None)

    spec_path = Path(spec_file)
    if not spec_path.exists():
        err_console.print('Unable to find spec file: {}'.format(spec_path.resolve()))
        raise typer.Exit(code=1)

    out_console.print('Processing spec...', end='\n\n', style='notify')
    with Progress(transient=True) as progress:
        parse_task = progress.add_task('Parsing config', total=100)
        entities = dsl_parser.parse_spec_file(spec_path)
        entities_map = {e.name: e for e in entities}
        progress.update(parse_task, advance=100)

        generate_task = progress.add_task('Generate classes', total=100)
        for entity in entities:
            progress.console.rule(entity.name)
            out_path = domain_generator.create_model_class(entity)
            progress.console.print(f'Domain model: {out_path}', style='bright_blue')

            out_path = domain_generator.create_command_class(entity, entities_map)
            progress.console.print(f'Domain command: {out_path}', style='bright_blue')

            out_path = domain_generator.create_db_port_class(entity)
            progress.console.print(f'DB port: {out_path}', style='bright_blue')

            out_path = domain_generator.create_use_case_port_class(entity)
            progress.console.print(f'Use case port: {out_path}', style='bright_blue')

            out_path = domain_generator.create_use_case_class(entity)
            progress.console.print(f'Use case: {out_path}', style='bright_blue')

            out_path = db_adapter_generator.create_entity_class(entity)
            progress.console.print(f'Entity: {out_path}', style='bright_cyan')

            out_path = db_adapter_generator.create_repository_class(entity)
            progress.console.print(f'Repository: {out_path}', style='bright_cyan')

            progress.update(generate_task, advance=100/len(entities))
            progress.console.print('')

        progress.update(generate_task, advance=100)

    out_console.print('\n-- All classes generated --\n', style='finished')


def _setup():
    theme = Theme({
        "notify": "cyan",
        "info": "dim cyan",
        "success": "green",
        "finished": "black on green",
        "warning": "yellow",
        "error": "red",
        "critical": "bold red"
    })

    global out_console
    out_console = Console(theme=theme)

    global err_console
    err_console = Console(stderr=True, style="red")



from pathlib import Path
from typing import List

from crudhex.domain.models import Entity
from crudhex.domain.utils.class_type_utils import get_import
from ..config_context import get_config
from . import command_generator
from . import model_generator
from . import use_case_port_generator
from . import db_port_generator

from crudhex.domain.ports import domain_code_writer

_PORT_PREFIX = 'UseCase'


def create_use_case_class(entity: Entity, folder: Path) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_use_case_type_name(entity)
    port_file = folder / f'{class_type}.java'

    id_type = entity.get_id_field().type.class_type
    model_type = model_generator.get_model_type_name(entity)
    command_type = command_generator.get_command_type_name(entity)
    db_port = db_port_generator.get_port_type_name(entity)

    domain_code_writer.create_use_case(port_file, class_type, get_package(),
                                       _get_use_case_imports(entity), id_type, model_type,
                                       command_type, command_type, db_port)

    return port_file


def get_package() -> str:
    return get_config().domain_use_cases_pkg


def get_use_case_type_name(entity: Entity) -> str:
    return f'{entity.name}{_PORT_PREFIX}'


def _get_use_case_imports(entity: Entity) -> List[str]:
    imports = use_case_port_generator._get_port_imports(entity)
    imports.extend([
        get_import(use_case_port_generator.get_package(), use_case_port_generator.get_port_type_name(entity)),
        get_import(db_port_generator.get_package(), db_port_generator.get_port_type_name(entity))

    ])

    return imports




from pathlib import Path
from typing import Optional, Dict

from crudhex.domain.models import Entity
from crudhex.domain.services.config_context import get_config
from crudhex.domain.services.domain import model_generator, command_generator, db_port_generator, use_case_port_generator, use_case_generator


def create_model_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_domain_models_path())

    return model_generator.create_model_class(entity, folder)


def create_command_class(entity: Entity, entities_map: Dict[str, Entity], folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_domain_commands_path())

    return command_generator.create_command_class(entity, entities_map, folder)


def create_db_port_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_domain_out_ports_path())

    return db_port_generator.create_port_class(entity, folder)


def create_use_case_port_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_domain_in_ports_path())

    return use_case_port_generator.create_port_class(entity, folder)


def create_use_case_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_domain_use_cases_path())

    return use_case_generator.create_use_case_class(entity, folder)



from typing import Dict, Optional, List

from crudhex.domain.models import Field
from crudhex.domain.utils import package_utils

_TYPE_RESOLVER: Optional['TypeResolver'] = None


def get_type_resolver(additional_types: Dict[str, str] = None) -> 'TypeResolver':
    global _TYPE_RESOLVER
    if not _TYPE_RESOLVER:
        _TYPE_RESOLVER = TypeResolver()

    if additional_types: _TYPE_RESOLVER.add_additional_types(additional_types)

    return _TYPE_RESOLVER


class TypeResolver:
    DEFAULT_COLLECTION = 'Set'
    DEFAULT_COLLECTION_IMPL = 'HashSet'
    GENERIC_TEMPLATE = '{}<{}>'

    _TYPE_DATA = {
        # Java
        'Collection': 'java.util.Collection',
        'List': 'java.util.List',
        'ArrayList': 'java.util.ArrayList',
        'Set': 'java.util.Set',
        'HashSet': 'java.util.HashSet',
        'Map': 'java.util.Map',
        'HashMap': 'java.util.HashMap',
        'UUID': 'java.util.UUID',
        'Optional': 'java.util.Optional',
        'Stream': 'java.util.stream.Stream',
        'String': 'java.lang.String',
        'Double': 'java.lang.Double',
        'Float': 'java.lang.Float',
        'Integer': 'java.lang.Integer',
        'Short': 'java.lang.Short',
        'Long': 'java.lang.Long',
        'Boolean': 'java.lang.Boolean',
        'BigDecimal': 'java.math.BigDecimal',
        'BigInteger': 'java.math.BigInteger',
        # Spring data
        'Page': 'org.springframework.data.domain.Page',
        'Pageable': 'org.springframework.data.domain.Pageable',

    }

    _COLLECTION_INTERFACES = [
        'Collection',
        'List',
        'Map',
        'Set'
    ]

    _COLLECTION_IMPL = {
        'Collection': 'ArrayList',
        'List': 'ArrayList',
        'Set': 'HashSet',
        'Map': 'HashMap'
    }

    type_data: Dict[str, str]

    def __init__(self, additional_types: Dict[str, str] = None):
        self.type_data = {}
        self._init()

        if additional_types:
            self.type_data.update(**additional_types)

    def resolve_type(self, class_type: str):
        return self.type_data.get(class_type, class_type)

    def add_additional_types(self, types: Dict[str, str]):
        self.type_data.update(**types)

    def get_field_types_full_class(self, field: Field, filter_native: bool = False) -> List[str]:
        if filter_native and field.type.is_native(): return []

        field_types = [field.type.get_qualified_class_type()]
        if not field.type.is_collection():
            return field_types

        field_types.append(self.resolve_type(field.type.collection_type))
        if self._is_collection_interface(field.type.collection_type):
            impl_type = self._get_collection_default_impl(field.type.collection_type)
            field_types.append(self.resolve_type(impl_type))

        return field_types

    def get_collection_type_impl(self, class_type: str) -> str:
        class_name = package_utils.get_class_name(class_type)

        # If not found in implementation map we assume is a collection implementation. Return same class then.
        if not self._is_collection_interface(class_name): return class_name

        return self._get_collection_default_impl(class_name)

    def _init(self):
        self.type_data.update(**TypeResolver._TYPE_DATA)

    def _is_collection_interface(self, class_type: str) -> bool:
        return class_type in TypeResolver._COLLECTION_INTERFACES

    def _get_collection_default_impl(self, class_type: str) -> str:
        return TypeResolver._COLLECTION_IMPL[class_type]



from pathlib import Path
from typing import Dict, List, Optional, Union

from crudhex.domain.models import RelationType
from crudhex.adapters.infrastructure.template_writer.config import template_config
from crudhex.adapters.infrastructure.template_writer.config.template_config import get_db_file_path
from crudhex.adapters.infrastructure.template_writer.services.template_env import get_template_environment


def create_entity(dest: Path, class_type: str, package: str,
                  imports: List[str], meta: Dict[str, str], fields: List[Dict[str, str]]):

    template_env = get_template_environment()

    id_field = _get_id_field(fields)
    if not id_field: raise RuntimeError('Id field is mandatory for db entity generation')

    fields_fragment = _generate_fields_fragment(fields)
    # TODO: Inverted relation sync accessors

    entity_template = template_env.get_template(get_db_file_path(template_config.DB_ENTITY_TEMPLATE))
    entity_code = entity_template.render({
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'fields': fields_fragment,
        'collection_methods': '',  # TODO
        'id_field': id_field['name'],
        **meta
    })

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(entity_code)


def create_entity_repository(dest: Path, class_type: str, package: str,
                             imports: List[str], entity_type: str, id_type: str):

    template_env = get_template_environment()

    repository_template = template_env.get_template(get_db_file_path(template_config.DB_REPOSITORY_TEMPLATE))
    repository_code = repository_template.render({
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'entity_type': entity_type,
        'id_type': id_type
    })

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(repository_code)


def _generate_fields_fragment(fields: List[Dict[str, Union[str, RelationType]]]) -> str:
    template_env = get_template_environment()

    templates = {}
    field_fragments = []
    for field in fields:
        if field['relationship']:
            template_name = template_config.get_relation_template(field['relationship'], field['mapped_by'] is None)
            template = templates.setdefault(
                template_name,
                template_env.get_template(get_db_file_path(template_name))
            )

            field_fragments.append(template.render(field))
        else:
            template = templates.setdefault(
                template_config.FIELD,
                template_env.get_template(get_db_file_path(template_config.FIELD))
            )

            field_fragments.append(template.render(field))

    return '\n\n'.join(field_fragments)


def _get_id_field(fields: List[Dict[str, str]]) -> Optional[Dict[str, str]]:
    id_field = None
    for field in fields:
        if field['id']:
            id_field = field
            break

    return id_field



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crudhex',
 'crudhex.adapters',
 'crudhex.adapters.application',
 'crudhex.adapters.application.cli',
 'crudhex.adapters.infrastructure',
 'crudhex.adapters.infrastructure.loader',
 'crudhex.adapters.infrastructure.template_writer',
 'crudhex.adapters.infrastructure.template_writer.adapters',
 'crudhex.adapters.infrastructure.template_writer.config',
 'crudhex.adapters.infrastructure.template_writer.services',
 'crudhex.domain',
 'crudhex.domain.config',
 'crudhex.domain.models',
 'crudhex.domain.ports',
 'crudhex.domain.services',
 'crudhex.domain.services.db',
 'crudhex.domain.services.domain',
 'crudhex.domain.utils']

package_data = \
{'': ['*'],
 'crudhex.adapters.infrastructure.template_writer': ['templates/db/*',
                                                     'templates/db/fragments/*',
                                                     'templates/domain/*',
                                                     'templates/domain/fragments/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0', 'PyYAML>=6.0,<7.0', 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['crudhex = crudhex.adapters.application.cli.main:main']}

setup_kwargs = {
    'name': 'crudhex',
    'version': '0.5.1',
    'description': 'Java/Spring CRUD code generator',
    'long_description': "# Crudhex\n\n[![PyPI version](https://badge.fury.io/py/crudhex.svg)](https://badge.fury.io/py/crudhex)\n\n⚠️ **Warn: Alpha development stage**\n\n---\n\nCLI tool to generate Java CRUD classes from a spec file. The target for this code generation is a Hexagonal architecture.\n\n\n## Motivation\nAdding CRUD operation in a hexagonal project is quite a pain. You can take shortcuts that can be totally legit in cases of just CRUD operations, but most cases if there is already a rich domain/application layer for other use cases CRUD shortcuts can break the consistency of the project.\n\nThe target of this CLI is to ease ~~my life~~ and try to give a general solution to CRUD generations in Hexagonal architecture. There are some customizations you can make, but some aspects for now are closed to customization for now.\n\n## Getting started\n\n### Project config file\nProject config file is used to know where things should go in the project. Usually contains a path to the sources folder (`src/main/java` in regular maven projects) and packages where things are located within that source folder.\nHere is a basic example:\n```yaml\nsrc: src/main/java # Java source folder for single module apps (where your packages start)\n\ndomain-models-pkg: com.salpreh.baseapi.domain.models # domain models package\ndb-models-pkg: com.salpreh.baseapi.adapers.infrasturcture.db.models # db entities package\nrest-models-pkg: com.salpreh.baseapi.adapters.api.models # rest api models\n```\n\nIn case of multi-module project you will have to specify `src` path to each module (domain, rest adapter and db adapter). More examples of config can be found in (`doc/examples/config`).\n\nThe default name for this config file is `.crudhex-conf.yaml`, located in the root of the project (you can provide the path to config file by cli options, so you can place it anywhere you want)\n\n### Spec file\nThis is the file where you specify the crud model. Class name, attributes and some meta data about DB structure (relations, PK field, column name alias, etc).\n\nAn example of spec file:\n```yaml\nPerson: # Model name\n  .meta:\n    table-name: persons # OPTIONAL: Table name for entity\n  id: # Field name\n    type: Long\n    id: sequence # PK marker. This field will be the primary key for the entity. \n                 # As value you specify generation strategy available in JPA with lower case.\n  birthPlanet: # Field name\n    type: Planet\n    column: birth_planet # OPTIONAL: Column name alias\n    relation: # OPTIONAL: DB relation meta data\n      type: many-to-one # Relation type\n      join-column: birth_planet_id # Join column for relation\n  affiliations:\n    type: Faction\n    relation:\n      type: many-to-many # In case of many-to-many we have a couple of more meta about DB setup\n      join-table: person_affiliation \n      join-column: person_id\n      inverse-join-column: faction_id\n  backup:\n    type: Person\n    relation:\n      type: one-to-one\n      join-column: backup_id\n  backing:\n    type: Person\n    relation:\n      type: one-to-one\n      mapped-by: backup # Mapping attribute for non-owning side of relation\n```\n\nSnippet config comments provide a overview of each relevant section in config. I'll dig further in config details in a dedicated section.\nFor now, you can find some aditional examples in `doc/examples/spec`.\n\n### Installation\n\nPackage is published in Pypi public repositories. You can use [pip]() (or another convinient python dependency manager) to install the package:\n```shell\npip install crudhex\n```\n\nOnce installed you should be able to use it as CLI tool:\n```shell\ncrudhex --help\n```",
    'author': 'salpreh',
    'author_email': 'salpreh.7@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/salpreh/crudhex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from collections import namedtuple
from pathlib import Path

from crudhex.domain.models import RelationType

# Pkg
TEMPLATES_PACKAGE = 'crudhex.adapters.infrastructure.template_writer'

# Base folders
TEMPLATE_FOLDER = 'templates'
DB_FOLDER = 'db'
DOMAIN_FOLDER = 'domain'
COMMONS_FOLDER = 'commons'
FRAGMENTS_FOLDER = 'fragments'

# DOMAIN TEMPLATES
MODEL_TEMPLATE = 'model.jinja2'
COMMAND_TEMPLATE = 'command.jinja2'
DB_PORT_TEMPLATE = 'db_port.jinja2'
USE_CASE_PORT_TEMPLATE = 'use_case_port.jinja2'
USE_CASE_TEMPLATE = 'use_case.jinja2'

# DOMAIN FRAGMENTS
DOM_FIELD = 'field.jinja2'

# DB TEMPLATES
DB_ENTITY_TEMPLATE = 'entity.jinja2'
DB_REPOSITORY_TEMPLATE = 'repository.jinja2'

# DB FRAGMENTS
COLLECTION_INVERSE_METHODS = 'collection_inverse_methods.jinja2'
COLLECTION_MAIN_METHODS = 'collection_main_methods.jinja2'
FIELD = 'field.jinja2'
ID_FIELD = 'id_field.jinja2'
M2M_INVERSE_FIELD = 'm2m_inverse_field.jinja2'
M2M_MAIN_FIELD = 'm2m_main_field.jinja2'
M2ONE_FIELD = 'm2one_field.jinja2'
ONE2M_FIELD = 'one2m_field.jinja2'
ONE2ONE_MAIN_FIELD = 'one2one_main_field.jinja2'
ONE2ONE_INVERSE_FIELD = 'one2one_inverse_field.jinja2'

_FRAGMENTS = [
    COLLECTION_INVERSE_METHODS,
    COLLECTION_MAIN_METHODS,
    FIELD,
    ID_FIELD,
    M2M_INVERSE_FIELD,
    M2M_MAIN_FIELD,
    M2ONE_FIELD,
    ONE2M_FIELD,
    ONE2ONE_MAIN_FIELD,
    ONE2ONE_INVERSE_FIELD,
    DOM_FIELD
]

RelationTemplate = namedtuple('RelationTemplate', 'main inverse')
_RELATION_TEMPLATE_MAP = {
    RelationType.ONE_TO_ONE: RelationTemplate(ONE2ONE_MAIN_FIELD, ONE2ONE_INVERSE_FIELD),
    RelationType.MANY_TO_ONE: RelationTemplate(M2ONE_FIELD, ONE2M_FIELD),
    RelationType.ONE_TO_MANY: RelationTemplate(M2ONE_FIELD, ONE2M_FIELD),  # ONE_TO_MANY always will be inverse side of relation. We keep it inverse side of template definition
    RelationType.MANY_TO_MANY: RelationTemplate(M2M_MAIN_FIELD, M2M_INVERSE_FIELD)
}


def get_db_file_path(file_name: str) -> str:
    file_path = _evaluate_file_type(Path(DB_FOLDER), file_name)

    return str(file_path / file_name)


def get_domain_file_path(file_name: str) -> str:
    file_path = _evaluate_file_type(Path(DOMAIN_FOLDER), file_name)

    return str(file_path / file_name)


def _evaluate_file_type(root_path: Path, file_name: str) -> Path:
    path = root_path
    if _is_fragment(file_name): path = root_path / FRAGMENTS_FOLDER

    return path


def _is_fragment(file_name: str) -> bool:
    return file_name in _FRAGMENTS


def get_relation_template(relation_type: RelationType, main: bool = True):
    relation_templates = _RELATION_TEMPLATE_MAP[relation_type]

    return relation_templates.main if main else relation_templates.inverse
